// -*-c++-*-
/* $Id: dupdb.C,v 1.5 1999/07/18 14:45:31 dm Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "dup.h"

void
dupdb::entry::init (const sketch::buf b)
{
  dupcount = 0;
  int i;
  for (i = 0; i < sketch::samplesize && b[i] != sketch::noshingle; i++)
    shingles[i] = b[i];
  entrysize = i;
  while (i < sketch::samplesize)
    shingles[i++] = 0;
}

u_int32_t
dupdb::entry::compare (const dupdb::entry &e) const
{
  u_int overlap = 0;
  int i = entrysize - 1, j = e.entrysize - 1;
  while (i >= 0 && j >= 0)
    if (shingles[i] < e.shingles[j])
      j--;
    else if (shingles[i] > e.shingles[j])
      i--;
    else {
      i--;
      j--;
      overlap++;
    }
  if (entrysize + e.entrysize == overlap)
    return sketch::max_resemblance;
  return ((overlap * (u_int64_t) sketch::max_resemblance)
	  / (entrysize + e.entrysize - overlap));
}

void
dupdb::init (u_int64_t p, u_int tp)
{
  assert (tp <= 100);
  assert (p < INT64 (0x200000000));
  magic = magic_val;
  poly = p;
  nentries = 0;
  if (!tp)
    tp = 90;
  threshold = tp * sketch::max_resemblance / 100;
}

bool
dupdb::check (size_t len) const
{
  if (len < offsetof (dupdb, entries))
    return false;
  if (magic != magic_val)
    return false;
  if (len < size ())
    return false;
  return true;
}

dupdb::entry *
dupdb::lookup (const entry &e)
{
  entry *ep = entries, *lim = ep + nentries;
  for (; ep < lim; ep++)
    if (e.compare (*ep) > threshold)
      return ep;
  return NULL;
}

void
dupfile::lock (bool shared)
{
  assert (fd >= 0);
  if (flock (fd, (update && !shared) ? LOCK_EX : LOCK_SH) < 0)
    fatal ("flock %s: %m\n", path);
}

void
dupfile::unlock ()
{
  msync ((caddr_t) dbp, dbsize, MS_ASYNC);
  if (flock (fd, LOCK_UN) < 0)
    fatal ("flock LOCK_UN %s: %m\n", path);
}

void
dupfile::unmap ()
{
  if (dbp) {
    if (munmap ((caddr_t) dbp, dbsize) < 0)
      fatal ("munmap: %m\n");
    dbp = NULL;
  }
}

void
dupfile::remap ()
{
  struct stat sb;
  if (fstat (fd, &sb) < 0)
    fatal ("fstat %s: %m\n", path);
  if (dbp) {
    if (implicit_cast<u_int64_t> (sb.st_size) == dbsize)
      return;
    unmap ();
  }
  dbsize = sb.st_size;
  if (dbsize) {
    dbp = (dupdb *) mmap (NULL, dbsize, PROT_READ | (update ? PROT_WRITE : 0),
			  MAP_FILE | MAP_SHARED, fd, 0);
    if (dbp == (dupdb *) MAP_FAILED)
      fatal ("mmap %s: %m\n", path);
  }
  else
    dbp = NULL;
}

dupfile::~dupfile ()
{
  if (fd >= 0) {
    unmap ();
    close (fd);
  }
  xfree (path);
  delete rp;
}

bool
dupfile::init (u_int tp = 0)
{
  assert (fd == -1);
  fd = open (path, update ? O_CREAT|O_RDWR : O_RDONLY, 0666);
  if (fd < 0)
    return false;

  lock ();

  struct stat sb;
  if (fstat (fd, &sb) < 0)
    fatal ("fstat %s: %m\n", path);
  dbsize = sb.st_size;
  if (!dbsize && update) {
    dupdb db;
    db.init (polygen (32), tp);
    dbsize = db.size ();
    if (implicit_cast<size_t> (write (fd, &db, dbsize)) != dbsize)
      fatal ("write %s: %m\n", path);
    created = true;
  }

  remap ();
  if (!dbp->check (dbsize)) {
    unlock ();
    unmap ();
    close (fd);
    return false;
  }
  rp = New rabinpoly (dbp->poly);

  unlock ();
  return true;
}

ssize_t
dupfile::lookup (const sketch::buf b)
{
  dupdb::entry e;
  e.init (b);

  lock ();
  remap ();
  dupdb::entry *ep = dbp->lookup (e);
  ssize_t ret = ep ? ep - dbp->entries : -1;

  if (!ep && update) {
    if (lseek (fd, dbp->size (), SEEK_SET) < 0)
      fatal ("lseek %s: %m\n", path);
    ret = dbp->nentries++;
    unmap ();
    if (write (fd, &e, sizeof (e)) != sizeof (e))
      fatal ("write %s: %m\n", path);
    remap ();
  }

  unlock ();
  return ret;
}

u_int32_t
dupfile::changecount (size_t entryno, int32_t inc)
{
  lock (!inc);
  remap ();
  assert (entryno < dbp->nentries);
  dupdb::entry *e = dbp->entries + entryno;
  u_int32_t ret = e->dupcount;
  if (inc < 0 && implicit_cast<u_int32_t> (-inc) >= ret)
    ret = 0;
  else
    ret += inc;
  if (inc)
    e->dupcount = ret;
  unlock ();
  return ret;
}
