/* $Id: sim.C,v 1.5 1999/07/15 23:18:31 dm Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "dup.h"
#include "itree.h"

struct result {
  str file1;
  str file2;
  u_int32_t sim;
  itree_entry<result> link;
  result (str f1, str f2, u_int32_t s) : file1 (f1), file2 (f2), sim (s) {}
};

typedef itree <u_int32_t, result, &result::sim, &result::link> restree;

char *opt_filter;

static int
runfilter (const rabinpoly &poly,
	   const char *filter, const char *file,
	   sketch::buf out)
{
  int fd = open (file, O_RDONLY);
  if (fd < 0) {
    warn ("%s: %m\n", file);
    return -1;
  }

  int fds[2];
  if (pipe (fds) < 0) {
    warn ("pipe: %m\n");
    close (fd);
    return -1;
  }

  pid_t pid = fork ();
  if (pid < 0) {
    warn ("fork: %m\n");
    return -1;
  }

  if (!pid) {
    if (fd != 0) {
      dup2 (fd, 0);
      close (fd);
    }
    if (fds[1] != 1) {
      dup2 (fds[1], 1);
      close (fds[1]);
    }
    close (fds[2]);
    execl ("/bin/sh", "/bin/sh", "-c", filter, NULL);
    fatal ("/bin/sh: %m\n");
  }

  close (fd);
  close (fds[1]);
  int res = sketchfd (poly, fds[0], out);
  close (fds[0]);
  int status;
  if (waitpid (pid, &status, 0) < 0)
    panic ("waitpid failed\n");
  if (status && !res) {
    warn ("%s: %s failed\n", file, filter);
    res = -1;
  }
  return res;
}

static void
usage ()
{
  warnx ("usage: %s [-f filter] file1 file2 ...\n", progname.cstr ());
  exit (1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "f:")) != -1)
    switch (ch) {
    case 'f':
      opt_filter = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;
  if (argc < 2)
    usage ();

  vector<str> files;
  rabinpoly rp (polygen (32));
  dupdb *ddb = static_cast<dupdb *> (malloc (dupdb::dbsize (argc)));
  ddb->init (rp.poly);

  for (int i = 0; i < argc; i++) {
    sketch::buf b;
    int res = opt_filter
      ? runfilter (rp, opt_filter, argv[i], b)
      : sketchfile (rp, argv[i], b);
    if (res >= 0) {
      files.push_back (argv[i]);
      ddb->entries[i].init (b);
    }
  }

  restree rt;
  for (u_int i = 0; i < files.size (); i++)
    for (u_int j = i + 1; j < files.size (); j++)
      rt.insert (New result (files[i], files[j],
			     ddb->entries[i].compare (ddb->entries[j])));

  for (result *r = rt.last (); r; r = rt.prev (r))
    printf ("%s %s %f\n", r->file1.cstr (), r->file2.cstr (),
	    r->sim / double (sketch::max_resemblance));

  return 0;
}
