/* $Id: sketch.C,v 1.5 1999/07/17 16:18:15 dm Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "dup.h"
#include "rxx.h"
#include <ctype.h>

const u_int64_t sketch::noshingle = (u_int64_t) -1;

void
sketch::addshingle (u_int64_t val)
{
  if (shingles[val])
    return;

  if (nshingles < samplesize)
    nshingles++;
  else if (val < maxshingle) {
    shingle *sp = shingles.last ();
    maxshingle = shingles.prev (sp)->val;
    delshingle (sp);
  }
  else
    return;
  
  if (val > maxshingle)
    maxshingle = val;
  shingles.insert (New shingle (val));
}

void
sketch::delshingle (sketch::shingle *sp)
{
  shingles.remove (sp);
  delete sp;
}

sketch::sketch (const rabinpoly &p)
  : poly (p), inword (false), wordno (1),
    nshingles (0), maxshingle (INT64 (0xffffffffffffffff))
{
  int i = shinglesize - 1;
  prints[i] = 1;
  while (i-- > 0)
    prints[i] = poly.append8 (prints[i+1], ' ');
}

sketch::~sketch ()
{
  shingles.traverse (wrap (this, &sketch::delshingle));
}

void
sketch::update (const void *_d, size_t len)
{
  const u_char *p = static_cast<const u_char *> (_d);
  const u_char *eom = p + len;

  if (!inword)
    goto dospace;

  for (;;) {
    for (;; p++) {
      if (p >= eom) {
	inword = true;
	return;
      }
      if (isspace (*p))
	break;
      for (int i = 0; i < shinglesize; i++)
	prints[i] = poly.append8 (prints[i], tolower (*p));
    }

    for (int i = 0; i < shinglesize; i++)
      prints[i] = poly.append8 (prints[i], ' ');
    /* Append zeros to ensure last byte affects all bits of fingerprint */
    for (int i = 0; i < 7; i++)
      prints[wordno] = poly.append8 (prints[wordno], 0);
    addshingle (prints[wordno]);
    prints[wordno] = 1;
    if (++wordno == shinglesize)
      wordno = 0;

  dospace:
    for (;; p++) {
      if (p >= eom) {
	inword = false;
	return;
      }
      if (!isspace (*p))
	break;
    }
  }
}

void
sketch::extract (sketch::buf out) const
{
  int i = 0;
  for (shingle *sp = shingles.first (); sp; sp = shingles.next (sp))
    out[i++] = sp->val;
  assert (i <= samplesize);
  while (i < samplesize)
    out[i++] = (u_int64_t) -1;
}

u_int32_t
sketch::resemblance (sketch::buf s1, sketch::buf s2)
{
  int i1 = 0, i2 = 0, isec = 0;
  while (i1 < samplesize && i2 < samplesize
	 && s1[i1] != (u_int64_t) -1 && s2[i2] != (u_int64_t) -1) {
    if (s1[i1] < s2[i2])
      i1++;
    else if (s2[i2] < s1[i1])
      i2++;
    else {
      i1++;
      i2++;
      isec++;
    }
  }
  while (i1 < samplesize && s1[i1] != (u_int64_t) -1) {
    i1++;
  }
  while (i2 < samplesize && s2[i2] != (u_int64_t) -1) {
    i2++;
  }
  if (i1 + i2 == isec)
    return max_resemblance;
  return (isec * (u_int64_t) max_resemblance) / (i1 + i2 - isec);
}

int
sketchfd (const rabinpoly &poly, int fd, sketch::buf out)
{
  sketch s (poly);
  char buf[8192];
  ssize_t n;
  while ((n = read (fd, buf, sizeof (buf))) > 0)
    s.update (buf, n);
  s.extract (out);
  return 0;
}

int
sketchfile (const rabinpoly &poly, const char *path, sketch::buf out)
{
  int fd = open (path, O_RDONLY);
  if (fd < 0) {
    warn ("%s: %m\n", path);
    return -1;
  }
  int ret = sketchfd (poly, fd, out);
  close (fd);
  return ret;
}

str
sketch822 (const rabinpoly &poly, int fd, sketch::buf b)
{
  static rxx usehdrs ("^(Subject:)", "i");
  static rxx getmsgid ("^Message-Id:\\s*(<[^\\s>]*>)", "i");
  sketch s (poly);
  suio in;
  str line;
  bool use = false;
  str msgid;

  for (;;) {
    while (!(line = suio_getline (&in)))
      if (in.input (fd, 8192) <= 0) {
	warnx ("incomplete message headers\n");
	exit (65);
      }
    if (!line.len ())
      break;
    if (!isspace (line[0]))
      use = usehdrs.match (line);
    if (use) {
      s.update (line.cstr (), line.len ());
      s.update (" ", 1);
    }
    if (getmsgid.match (line))
      msgid = getmsgid[1];
  }

  if (!msgid) {
    warnx ("could not parse Message-ID\n");
    exit (65);
  }

  for (const iovec *iov = in.iov (); iov < in.iovlim (); iov++)
    s.update (iov->iov_base, iov->iov_len);
  char buf[8192];
  ssize_t n;
  while ((n = read (fd, buf, sizeof (buf))) > 0)
    s.update (buf, n);

  s.extract (b);
  return msgid;
}
